from prophet.plot import plot_plotly, plot_components_plotly
from prophet import Prophet
from cassandra.references.load_holidays import load_holidays
import pandas as pd


def prophet(df, name_date_column, name_target_column, window_start='', window_end='',
            national_holidays_abbreviation='IT', future_dataframe_periods=14, plot_prophet=True,
            weekly_seasonality=True, yearly_seasonality=True, daily_seasonality=True, seasonality_mode='additive'):
    # Read the CSV on holidays
    csv_holiday = load_holidays()

    # Select the holidays according to the country that interests me
    condition = (csv_holiday['country'] == national_holidays_abbreviation)
    holidays = csv_holiday.loc[condition, ['ds', 'holiday']]

    # Create a DF with the only two columns for Prophet
    prophet_df = df[[name_date_column, name_target_column]]

    # Rename the columns for Prophet
    prophet_df = prophet_df.rename(columns={name_date_column: 'ds', name_target_column: 'y'})

    # Instance and fit Prophet
    prophet_m = Prophet(weekly_seasonality=weekly_seasonality, yearly_seasonality=yearly_seasonality,
                        daily_seasonality=daily_seasonality, holidays=holidays, seasonality_mode=seasonality_mode)
    prophet_m.add_seasonality(name='monthly', period=30.5, fourier_order=5)
    prophet_m.fit(prophet_df)

    future = prophet_m.make_future_dataframe(periods=future_dataframe_periods)

    forecast = prophet_m.predict(future)

    new_forecast = forecast[['ds', 'yhat', 'trend', 'holidays', 'additive_terms', 'multiplicative_terms']].copy()

    if 'yearly' in forecast:
        new_forecast['yearly'] = forecast['yearly'].copy()
    if 'monthly' in forecast:
        new_forecast['monthly'] = forecast['monthly'].copy()
    if 'weekly' in forecast:
        new_forecast['weekly'] = forecast['weekly'].copy()
    if 'daily' in forecast:
        new_forecast['daily'] = forecast['daily'].copy()

    if plot_prophet:
        prophet_m.plot(forecast)
        prophet_m.plot_components(forecast)

    # plot_components_plotly(prophet_m, forecast)

    sub_prophet_df = new_forecast[['ds', 'trend', 'holidays']].copy()

    if 'yearly' in new_forecast:
        sub_prophet_df['yearly'] = new_forecast['yearly']
    if 'monthly' in forecast:
        sub_prophet_df['monthly'] = new_forecast['monthly']
    if 'weekly' in forecast:
        sub_prophet_df['weekly'] = new_forecast['weekly']
    if 'daily' in forecast:
        sub_prophet_df['daily'] = new_forecast['daily']

    sub_prophet_df = sub_prophet_df.rename(columns={'ds': name_date_column})

    df[name_date_column] = pd.to_datetime(df[name_date_column])
    sub_prophet_df[name_date_column] = pd.to_datetime(sub_prophet_df[name_date_column])

    full_df = pd.merge(df, sub_prophet_df, how='inner', on=name_date_column)

    if window_start:
        full_df.drop(full_df[full_df[name_date_column] < window_start].index, inplace=True)
    if window_end:
        full_df.drop(full_df[full_df[name_date_column] > window_end].index, inplace=True)

    return full_df



#!/usr/bin/env python
# coding: utf-8

# In[1]:


# Imports
import pandas as pd
import numpy as np
import nevergrad as ng

from sklearn.linear_model import Ridge
from sklearn.linear_model import LinearRegression
from sklearn.model_selection import train_test_split
from sklearn.neural_network import MLPRegressor

from src.cassandra.model.linear import linear
from src.cassandra.model.logLinear import logLinear
from src.cassandra.model.logLog import logLog
from src.cassandra.model.mlpRegressor import mlpRegressor
from src.cassandra.model.ridge import ridge
from src.cassandra.model.modelEvaluation.plot import show_nrmse, show_mape, show_rsquared, show_coefficients
from src.cassandra.model.nestedModel import nestedModel, mainNestedModel
from src.cassandra.data.dataProcessing.prophet import prophet
from src.cassandra.data.trasformations.trasformations import saturation, adstock, adstock_weibull
from src.cassandra.model.modelEvaluation.evaluation import response_regression_to_dataset, adstock_saturation_to_dataset, \
    decomposition_to_dataset, spend_to_dataset
from src.cassandra.data.dataAnalysis.plot import show_target_vs_media_spent_graph, show_prediction_vs_actual_graph, \
    show_mape_trend, show_prediction_vs_actual_graph_with_error
from src.cassandra.data.dataProcessing.clean_and_format import create_dataset, create_data_columns
from src.cassandra.data.featureSelection.ols import ols
from src.cassandra.budgetAllocator.budget_allocator import *
from src.cassandra.budgetAllocator.utils import *
from src.cassandra.utils.save import save_metric_to_csv, save_timeseries_to_csv, save_decomposition_to_csv, \
    save_dataset_to_csv

# In[2]:


# pip install cassandra-mmm --upgrade


# In[25]:


subset_columns = ['date', 'conversions', 'fb_pr_spend', 'fb_re_spend',
                  'g_discovery_spend', 'g_display_spend', 'g_performance_max_spend', 'g_search_spend', 'g_video_spend',
                  'tiktok_learn_spend', 'influencer_spend', 'email_campaigns_sessions',
                  's_organic', 's_referral', 's_influencer', 's_affiliate', 's_direct', 's_partner', 's_mediapartner',
                  'adopta_albero_impressions', 'regalar_arbol_impressions', 'regala_arbol_impressions',
                  'plantar_arbol_impressions', 'treedom_impressions',
                  # 'avg_p_cacao', 'avg_p_baobab', 'avg_p_asc2020', 'avg_p_pt2020', 'avg_p_ft_ag',
                  'tickets_count']
csv_data = create_dataset('D:/workspace_cassandra/mmm/treedom/ES/treedom-es - modifiche.csv', subset_columns=subset_columns)

# csv_data = create_data_columns(csv_data, 'date', national_holidays_abbreviation = 'ES')

# csv_data.drop(csv_data[csv_data['date'] < '2021-02-28'].index, inplace=True)
# csv_data.drop(csv_data[csv_data['conversions'] < 1].index, inplace=True)
# csv_data.drop(csv_data[csv_data['conversions'] > 55].index, inplace=True)

# pd.set_option("display.max_rows", None)
# pd.set_option("display.max_columns", None)
pd.set_option("display.max_rows", 20)
pd.set_option("display.max_columns", 20)
csv_data.fillna(method='bfill', inplace=True)
csv_data.fillna(method='ffill', inplace=True)
csv_data.fillna(0, inplace=True)
csv_data

# # Data Preparation

# In[26]:


keyword_cols_aggregate = ['adopta_albero_impressions', 'regalar_arbol_impressions', 'regala_arbol_impressions',
                          'plantar_arbol_impressions', 'treedom_impressions']
csv_data['keywords_impressions'] = csv_data[keyword_cols_aggregate].sum(axis=1)

csv_data.drop(keyword_cols_aggregate, axis=1, inplace=True)

s_mediapartner_aggregate = ['s_mediapartner', 's_partner']
csv_data['s_partners'] = csv_data[s_mediapartner_aggregate].sum(axis=1)

csv_data.drop(s_mediapartner_aggregate, axis=1, inplace=True)

s_influencer_aggregate = ['s_influencer', 's_affiliate']
csv_data['s_influencers'] = csv_data[s_influencer_aggregate].sum(axis=1)

csv_data.drop(s_influencer_aggregate, axis=1, inplace=True)

s_other_aggregate = ['email_campaigns_sessions', 's_partners']
csv_data['s_others'] = csv_data[s_other_aggregate].sum(axis=1)

csv_data.drop(s_other_aggregate, axis=1, inplace=True)

# products_vars = [col for col in df.columns if 'avg' in col]

# for p in products_vars:
# for index, row in csv_data.iterrows():
# if row[p] != 0:
# csv_data.at[index, p] = 1
# else:
# csv_data.at[index, p] = 0


# # Prophet

# In[27]:


df = prophet(csv_data, 'date', 'conversions', window_start='2021-02-28', national_holidays_abbreviation='ES',
             future_dataframe_periods=365, plot_prophet=True, daily_seasonality=False)

# In[28]:


df

# # Spends

# In[7]:


medias_google = [col for col in df.columns if 'spend' in col if 'g_' in col]
medias_fb = [col for col in df.columns if 'spend' in col if 'fb_' in col]
medias_others = [col for col in df.columns if 'spend' in col if 'fb_' not in col if 'g_' not in col]

media = medias_google + medias_fb + medias_others

spends_df = spend_to_dataset(df, media)
spends_df

# In[8]:


spends, response_get_budget = getBudget(df, 'date', 'conversions', media, '2022-02-28')
print(spends, np.sum(spends), response_get_budget)

# # Correlation

# In[29]:


pd.set_option("display.max_rows", None)
pd.set_option("display.max_columns", None)
df.corr()

# In[51]:


organic = ['s_organic', 's_referral', 's_influencers', 's_direct', 's_others', 'keywords_impressions']
context = ['tickets_count']

# product_vars = [col for col in df.columns if 'avg' in col]

prophet_vars = ['trend', 'yearly', 'monthly', 'weekly', 'holidays']
# prophet_vars = ['day_week', 'month', 'festivo']

all_features = media + organic + context + prophet_vars  # + product_vars

# In[52]:


X = df[all_features]
y = df['conversions']

result_ols, model_ols = ols(X, y, constant=True)

# # Nested Model

# In[32]:


# Sub-model describes Keywords Impressions using Google and Facebook's Media Spents
# Mail model uses all other variables and the sub-model to describe conversions
# medias_main = medias_google + medias_fb

# result, model, metrics_values, sub_result, sub_model, sub_metric_values = mainNestedModel(df, medias_main, 'keywords_impressions', all_variables, 'conversions', sub_model_regression = 'linear', sub_metric = ['rsq', 'nrmse', 'mape'], return_sub_metric = True, model_regression = 'ridge', metric = ['rsq', 'nrmse', 'mape'], return_metric = True)


# # Ridge Regression

# In[53]:


result_ridge, model_ridge = ridge(df, X.columns, 'conversions', 'Ridge', ridge_number=21,
                                  metric=['rsq', 'nrmse', 'mape'])

# In[55]:


show_prediction_vs_actual_graph_with_error(result_ridge, result_ridge['conversions'], result_ridge['prediction'],
                                           'date', 'revenue', name_model='')

# In[56]:


show_coefficients(all_features, model_ridge, 'Ridge')

# # Linear Regression

# In[36]:


#result_linear, model_linear = linear(df, X.columns, 'conversions', 'Linear', metric=['rsq', 'nrmse', 'mape'])

# In[37]:


#show_prediction_vs_actual_graph_with_error(result_linear, result_linear['conversions'], result_linear['prediction'],
                                           #'date', 'conversions', name_model='')

# In[38]:


#show_coefficients(all_features, model_linear, 'Linear')

# # LogLog Linear Regression

# In[39]:


trasformations_log_columns = media + organic
not_log_columns = prophet_vars + context

# In[40]:


#result_logLog_linear, model_logLog_linear = logLog(df, trasformations_log_columns, not_log_columns, 'conversions',
                                                   #'logLog', model_regression='linear', metric=['rsq', 'nrmse', 'mape'])

# In[41]:


# show_prediction_vs_actual_graph(result_logLog, 'date', 'conversions', 'prediction')


# In[42]:


#show_prediction_vs_actual_graph_with_error(result_logLog_linear, result_logLog_linear['conversions'],
                                           #result_logLog_linear['prediction'], 'date', 'conversions', name_model='')

# In[43]:


#show_coefficients(all_features, model_logLog_linear, 'Log Log')

# # LogLog Ridge Regression

# In[44]:


result_logLog_ridge, model_logLog_ridge = logLog(df, trasformations_log_columns, not_log_columns, 'conversions',
                                                 'logLog', model_regression='ridge', metric=['rsq', 'nrmse', 'mape'])

# In[45]:


show_prediction_vs_actual_graph_with_error(result_logLog_ridge, result_logLog_ridge['conversions'],
                                           result_logLog_ridge['prediction'], 'date', 'conversions', name_model='')

# In[46]:


show_coefficients(all_features, model_logLog_ridge, 'Log Log Ridge')

# # Modello Ridge

# In[47]:


mape_array = []
df_transf = df.copy()


def build_model(b1, b2, b3, b4, b5, b6, b7, b8, b11, b13, b14, b15, b16, b17, b18, b19,
                b20, b21, b22, b23, b24, b0,  # b26, b27, b28, b29, b30, b31, b32, b33,
                g_discovery_spend_scale, g_discovery_spend_shape, g_discovery_spend_beta,
                g_display_spend_scale, g_display_spend_shape, g_display_spend_beta,
                g_performance_max_spend_scale, g_performance_max_spend_shape, g_performance_max_spend_beta,
                g_search_spend_scale, g_search_spend_shape, g_search_spend_beta,
                g_video_spend_scale, g_video_spend_shape, g_video_spend_beta,
                fb_re_spend_scale, fb_re_spend_shape, fb_re_spend_beta,
                fb_pr_spend_scale, fb_pr_spend_shape, fb_pr_spend_beta,
                tiktok_learn_spend_scale, tiktok_learn_spend_shape, tiktok_learn_spend_beta,
                # microsoft_spend_scale, microsoft_spend_shape, microsoft_spend_beta,
                # linkedin_spend_scale, linkedin_spend_shape, linkedin_spend_beta,
                influencer_spend_scale, influencer_spend_shape, influencer_spend_beta):
    # Google
    x1 = df['g_discovery_spend']
    x2 = df['g_display_spend']
    x3 = df['g_performance_max_spend']
    x4 = df['g_search_spend']
    x5 = df['g_video_spend']

    x1_adstock = adstock_weibull(x1, g_discovery_spend_scale, g_discovery_spend_shape).replace([np.inf, -np.inf], 0)
    x2_adstock = adstock_weibull(x2, g_display_spend_scale, g_display_spend_shape).replace([np.inf, -np.inf], 0)
    x3_adstock = adstock_weibull(x3, g_performance_max_spend_scale, g_performance_max_spend_shape).replace(
        [np.inf, -np.inf], 0)
    x4_adstock = adstock_weibull(x4, g_search_spend_scale, g_search_spend_shape).replace([np.inf, -np.inf], 0)
    x5_adstock = adstock_weibull(x5, g_video_spend_scale, g_video_spend_shape).replace([np.inf, -np.inf], 0)

    x1 = saturation(x1_adstock, g_discovery_spend_beta)
    x2 = saturation(x2_adstock, g_display_spend_beta)
    x3 = saturation(x3_adstock, g_performance_max_spend_beta)
    x4 = saturation(x4_adstock, g_search_spend_beta)
    x5 = saturation(x5_adstock, g_video_spend_beta)

    # Facebook
    x6 = df['fb_re_spend']
    x7 = df['fb_pr_spend']

    x6_adstock = adstock_weibull(x6, fb_re_spend_scale, fb_re_spend_shape).replace([np.inf, -np.inf], 0)
    x7_adstock = adstock_weibull(x7, fb_pr_spend_scale, fb_pr_spend_shape).replace([np.inf, -np.inf], 0)

    x6 = saturation(x6_adstock, fb_re_spend_beta)
    x7 = saturation(x7_adstock, fb_pr_spend_beta)

    # Other Medias
    x8 = df['tiktok_learn_spend']
    # x9 = df['microsoft_spend']
    # x10 = df['linkedin_spend']
    x11 = df['influencer_spend']

    x8_adstock = adstock_weibull(x8, tiktok_learn_spend_scale, tiktok_learn_spend_shape).replace([np.inf, -np.inf], 0)
    # x9_adstock = adstock_weibull(x9, microsoft_spend_scale, microsoft_spend_shape).replace([np.inf, -np.inf], 0)
    # x10_adstock = adstock_weibull(x10, linkedin_spend_scale, linkedin_spend_shape).replace([np.inf, -np.inf], 0)
    x11_adstock = adstock_weibull(x11, influencer_spend_scale, influencer_spend_shape).replace([np.inf, -np.inf], 0)

    x8 = saturation(x8_adstock, tiktok_learn_spend_beta)
    # x9 = saturation(x9_adstock, microsoft_spend_beta)
    # x10 = saturation(x10_adstock, linkedin_spend_beta)
    x11 = saturation(x11_adstock, influencer_spend_beta)

    # Google
    df_transf['g_discovery_spend'] = x1
    df_transf['g_display_spend'] = x2
    df_transf['g_performance_max_spend'] = x3
    df_transf['g_search_spend'] = x4
    df_transf['g_video_spend'] = x5

    # Facebook
    df_transf['fb_re_spend'] = x6
    df_transf['fb_pr_spend'] = x7

    # Other Medias
    df_transf['tiktok_learn_spend'] = x8
    # df_transf['microsoft_spend'] = x9
    # df_transf['linkedin_spend'] = x10
    df_transf['influencer_spend'] = x11

    X = df_transf[all_features]
    y = df_transf['conversions']

    X_train, X_test, y_train, y_test = train_test_split(X, y)

    model = Ridge(22)

    model.fit(X_train, y_train)

    coefs = [b1] + [b2] + [b3] + [b4] + [b5] + [b6] + [b7] + [b8] + [b11] + [b13] + [b14] + [b15] + [b16] + [b17] + [
        b18] + [b19] + [b20] + [b21] + [b22] + [b23] + [b24]
    model.coef_ = np.array(coefs)
    model.intercept_ = b0

    # Setting the Result Dataset
    result = df_transf.copy()

    result['prediction'] = model.predict(X)

    nrmse_val = show_nrmse(result['conversions'], result['prediction'])
    mape_val = show_mape(result['conversions'], result['prediction'])
    rsquared_val = show_rsquared(result['conversions'], result['prediction'])

    print(rsquared_val, nrmse_val, mape_val)

    mape_array.append(mape_val)

    return mape_val


# In[48]:


instrum = ng.p.Instrumentation(
    # alpha = ng.p.Scalar(lower=0),

    b1=ng.p.Scalar(lower=0),
    b2=ng.p.Scalar(lower=0),
    b3=ng.p.Scalar(lower=0),
    b4=ng.p.Scalar(lower=0),
    b5=ng.p.Scalar(lower=0),
    b6=ng.p.Scalar(lower=0),
    b7=ng.p.Scalar(lower=0),
    b8=ng.p.Scalar(lower=0),
    # b9 = ng.p.Scalar(lower=0),
    # b10 = ng.p.Scalar(lower=0),
    b11=ng.p.Scalar(lower=0),

    # b12 = ng.p.Scalar(),
    b13=ng.p.Scalar(lower=0),
    b14=ng.p.Scalar(lower=0),
    b15=ng.p.Scalar(lower=0),
    b16=ng.p.Scalar(lower=0),
    b17=ng.p.Scalar(lower=0),
    b18=ng.p.Scalar(lower=0),
    b19=ng.p.Scalar(upper=0),
    b20=ng.p.Scalar(),
    b21=ng.p.Scalar(),
    b22=ng.p.Scalar(),
    b23=ng.p.Scalar(),
    b24=ng.p.Scalar(),
    # b25 = ng.p.Scalar(),
    # b26 = ng.p.Scalar(),
    # b27 = ng.p.Scalar(),
    # b28 = ng.p.Scalar(),
    # b29 = ng.p.Scalar(),
    # b30 = ng.p.Scalar(),
    # b31 = ng.p.Scalar(),
    # b32 = ng.p.Scalar(),
    # b33 = ng.p.Scalar(),

    b0=ng.p.Scalar(),

    g_discovery_spend_scale=ng.p.Scalar(lower=0, upper=0.1),
    g_discovery_spend_shape=ng.p.Scalar(lower=0, upper=2),
    g_discovery_spend_beta=ng.p.Scalar(lower=0, upper=1),
    g_display_spend_scale=ng.p.Scalar(lower=0, upper=0.1),
    g_display_spend_shape=ng.p.Scalar(lower=0, upper=2),
    g_display_spend_beta=ng.p.Scalar(lower=0, upper=1),
    g_performance_max_spend_scale=ng.p.Scalar(lower=0, upper=0.1),
    g_performance_max_spend_shape=ng.p.Scalar(lower=0, upper=2),
    g_performance_max_spend_beta=ng.p.Scalar(lower=0, upper=1),
    g_search_spend_scale=ng.p.Scalar(lower=0, upper=0.1),
    g_search_spend_shape=ng.p.Scalar(lower=0, upper=2),
    g_search_spend_beta=ng.p.Scalar(lower=0, upper=1),
    g_video_spend_scale=ng.p.Scalar(lower=0, upper=0.1),
    g_video_spend_shape=ng.p.Scalar(lower=0, upper=2),
    g_video_spend_beta=ng.p.Scalar(lower=0, upper=1),

    fb_re_spend_scale=ng.p.Scalar(lower=0, upper=0.1),
    fb_re_spend_shape=ng.p.Scalar(lower=0, upper=2),
    fb_re_spend_beta=ng.p.Scalar(lower=0, upper=1),
    fb_pr_spend_scale=ng.p.Scalar(lower=0, upper=0.1),
    fb_pr_spend_shape=ng.p.Scalar(lower=0, upper=2),
    fb_pr_spend_beta=ng.p.Scalar(lower=0, upper=1),

    tiktok_learn_spend_scale=ng.p.Scalar(lower=0, upper=0.1),
    tiktok_learn_spend_shape=ng.p.Scalar(lower=0, upper=2),
    tiktok_learn_spend_beta=ng.p.Scalar(lower=0, upper=1),

    # microsoft_spend_scale = ng.p.Scalar(lower=0, upper=0.1),
    # microsoft_spend_shape = ng.p.Scalar(lower=0, upper=2),
    # microsoft_spend_beta = ng.p.Scalar(lower=0, upper=1),

    # linkedin_spend_scale = ng.p.Scalar(lower=0, upper=0.1),
    # linkedin_spend_shape = ng.p.Scalar(lower=0, upper=2),
    # linkedin_spend_beta = ng.p.Scalar(lower=0, upper=1),

    influencer_spend_scale=ng.p.Scalar(lower=0, upper=0.1),
    influencer_spend_shape=ng.p.Scalar(lower=0, upper=2),
    influencer_spend_beta=ng.p.Scalar(lower=0, upper=1)
)

optimizer = ng.optimizers.TwoPointsDE(parametrization=instrum, budget=1000)
recommendation = optimizer.minimize(build_model)

# NGOpt
# TwoPointsDE
recommendation.value


# In[49]:


def build_model(b1, b2, b3, b4, b5, b6, b7, b8, b11, b13, b14, b15, b16, b17, b18, b19,
                b20, b21, b22, b23, b24, b0,  # b26, b27, b28, b29, b30, b31, b32, b33,
                g_discovery_spend_scale, g_discovery_spend_shape, g_discovery_spend_beta,
                g_display_spend_scale, g_display_spend_shape, g_display_spend_beta,
                g_performance_max_spend_scale, g_performance_max_spend_shape, g_performance_max_spend_beta,
                g_search_spend_scale, g_search_spend_shape, g_search_spend_beta,
                g_video_spend_scale, g_video_spend_shape, g_video_spend_beta,
                fb_re_spend_scale, fb_re_spend_shape, fb_re_spend_beta,
                fb_pr_spend_scale, fb_pr_spend_shape, fb_pr_spend_beta,
                tiktok_learn_spend_scale, tiktok_learn_spend_shape, tiktok_learn_spend_beta,
                # microsoft_spend_scale, microsoft_spend_shape, microsoft_spend_beta,
                # linkedin_spend_scale, linkedin_spend_shape, linkedin_spend_beta,
                influencer_spend_scale, influencer_spend_shape, influencer_spend_beta):
    # Google
    x1 = df['g_discovery_spend']
    x2 = df['g_display_spend']
    x3 = df['g_performance_max_spend']
    x4 = df['g_search_spend']
    x5 = df['g_video_spend']

    x1_adstock = adstock_weibull(x1, g_discovery_spend_scale, g_discovery_spend_shape).replace([np.inf, -np.inf], 0)
    x2_adstock = adstock_weibull(x2, g_display_spend_scale, g_display_spend_shape).replace([np.inf, -np.inf], 0)
    x3_adstock = adstock_weibull(x3, g_performance_max_spend_scale, g_performance_max_spend_shape).replace(
        [np.inf, -np.inf], 0)
    x4_adstock = adstock_weibull(x4, g_search_spend_scale, g_search_spend_shape).replace([np.inf, -np.inf], 0)
    x5_adstock = adstock_weibull(x5, g_video_spend_scale, g_video_spend_shape).replace([np.inf, -np.inf], 0)

    x1 = saturation(x1_adstock, g_discovery_spend_beta)
    x2 = saturation(x2_adstock, g_display_spend_beta)
    x3 = saturation(x3_adstock, g_performance_max_spend_beta)
    x4 = saturation(x4_adstock, g_search_spend_beta)
    x5 = saturation(x5_adstock, g_video_spend_beta)

    # Facebook
    x6 = df['fb_re_spend']
    x7 = df['fb_pr_spend']

    x6_adstock = adstock_weibull(x6, fb_re_spend_scale, fb_re_spend_shape).replace([np.inf, -np.inf], 0)
    x7_adstock = adstock_weibull(x7, fb_pr_spend_scale, fb_pr_spend_shape).replace([np.inf, -np.inf], 0)

    x6 = saturation(x6_adstock, fb_re_spend_beta)
    x7 = saturation(x7_adstock, fb_pr_spend_beta)

    # Other Medias
    x8 = df['tiktok_learn_spend']
    # x9 = df['microsoft_spend']
    # x10 = df['linkedin_spend']
    x11 = df['influencer_spend']

    x8_adstock = adstock_weibull(x8, tiktok_learn_spend_scale, tiktok_learn_spend_shape).replace([np.inf, -np.inf], 0)
    # x9_adstock = adstock_weibull(x9, microsoft_spend_scale, microsoft_spend_shape).replace([np.inf, -np.inf], 0)
    # x10_adstock = adstock_weibull(x10, linkedin_spend_scale, linkedin_spend_shape).replace([np.inf, -np.inf], 0)
    x11_adstock = adstock_weibull(x11, influencer_spend_scale, influencer_spend_shape).replace([np.inf, -np.inf], 0)

    x8 = saturation(x8_adstock, tiktok_learn_spend_beta)
    # x9 = saturation(x9_adstock, microsoft_spend_beta)
    # x10 = saturation(x10_adstock, linkedin_spend_beta)
    x11 = saturation(x11_adstock, influencer_spend_beta)

    # Google
    df_transf['g_discovery_spend'] = x1
    df_transf['g_display_spend'] = x2
    df_transf['g_performance_max_spend'] = x3
    df_transf['g_search_spend'] = x4
    df_transf['g_video_spend'] = x5

    # Facebook
    df_transf['fb_re_spend'] = x6
    df_transf['fb_pr_spend'] = x7

    # Other Medias
    df_transf['tiktok_learn_spend'] = x8
    # df_transf['microsoft_spend'] = x9
    # df_transf['linkedin_spend'] = x10
    df_transf['influencer_spend'] = x11

    X = df_transf[all_features]
    y = df_transf['conversions']

    X_train, X_test, y_train, y_test = train_test_split(X, y)

    model = Ridge(22)

    model.fit(X_train, y_train)

    coefs = [b1] + [b2] + [b3] + [b4] + [b5] + [b6] + [b7] + [b8] + [b11] + [b13] + [b14] + [b15] + [b16] + [b17] + [
        b18] + [b19] + [b20] + [b21] + [b22] + [b23] + [b24]
    model.coef_ = np.array(coefs)
    model.intercept_ = b0

    # Setting the Result Dataset
    result = df_transf.copy()

    result['prediction'] = model.predict(X)

    nrmse_val = show_nrmse(result['conversions'], result['prediction'])
    mape_val = show_mape(result['conversions'], result['prediction'])
    rsquared_val = show_rsquared(result['conversions'], result['prediction'])

    print('RSQUARED: ', rsquared_val)
    print('NRMSE: ', nrmse_val)
    print('MAPE: ', mape_val)

    return result, model


result, model = build_model(**recommendation.value[1])

# In[24]:


show_mape_trend(mape_array)

# In[41]:


metric = {}
metric['rsq'] = 0.897
metric['nrmse'] = 0.413
metric['mape'] = 0.239

# In[42]:


save_metric_to_csv(metric, 'C:/Users/Federico/Downloads/KPIs.csv')

# In[44]:


coefficenti = show_coefficients(all_features, model, 'Ridge', graph=True)

# In[23]:


show_prediction_vs_actual_graph(result, 'date', 'conversions', 'prediction')

# In[24]:


pd.set_option("display.max_rows", 20)
pd.set_option("display.max_columns", 20)
result

# In[25]:


df_adstock_saturation, result_df_adstock = adstock_saturation_to_dataset(recommendation.value[1], 'weibull')
save_dataset_to_csv(df_adstock_saturation, 'C:/Users/Federico/Downloads/treddom_ES_Adstock_Saturation.csv')
df_adstock_saturation

# In[27]:


coefficenti_dict = {}
for index, value in enumerate(coefficenti):
    coefficenti_dict[index] = value
coefficenti_dict['b0'] = model.intercept_

# In[67]:


# coefficenti = recommendation.value[1].copy()
# del coefficenti['alpha']
columns_coef = ['g_discovery_spend', 'g_display_spend', 'g_performance_max_spend', 'g_search_spend', 'g_video_spend',
                'fb_pr_spend', 'fb_re_spend', 'tiktok_learn_spend', 'influencer_spend',
                's_organic', 's_referral', 's_influencers', 's_direct', 's_others', 'keywords_impressions',
                # 'avg_p_cacao', 'avg_p_baobab', 'avg_p_asc2020', 'avg_p_pt2020', 'avg_p_ft_ag',
                'tickets_count', 'trend', 'yearly', 'monthly', 'weekly', 'holidays']

# In[73]:


df_saturation = df.copy()

for index, row in df_saturation.iterrows():
    for m in list(df_adstock_saturation['canale']):
        df_saturation.at[index, m] = saturation(df_saturation.at[index, m], df_adstock_saturation.loc[
            df_adstock_saturation['canale'] == m, 'saturation'].iloc[0])

coef_array = coefficenti
new_dict = {}

for x in range(len(coef_array) - 1):
    new_dict[columns_coef[x]] = df_saturation[columns_coef[x]] * coef_array[x]

new_dict['intercept'] = coef_array[-1]

pd.set_option("display.max_rows", 20)
pd.set_option("display.max_columns", 20)
df_saturation_response = pd.DataFrame(new_dict)

# In[144]:


# s_df = pd.merge(df['g_discovery_spend'], df_saturation_response['g_discovery_spend'], left_index=True, right_index=True)
# pd.set_option("display.max_rows", None)
# pd.set_option("display.max_columns", None)
# s_df


# In[145]:


import plotly.express as px

for m in media:
    s_df = pd.merge(df[m], df_saturation_response[m], left_index=True, right_index=True)

    # For each media remove all the 0 spend values as Log(0) can't be calculated
    s_df[m + '_x'] = s_df[m + '_x'][s_df[m + '_x'] != 0]
    s_df[m + '_y'] = s_df[m + '_y'][s_df[m + '_y'] != 0]
    # saturation_df = saturation_df[(saturation_df[[media]] != 0).all(axis=1)]

    # Create a Scatter Plot
    fig = px.scatter(
        # Our temporary DF with no 0 values
        s_df,
        # Media channel's name
        x=m + '_x',
        # Our output variable (Sales or Conversion)
        y=m + '_y',
        # Specify the type of trendline you want to show, use OLS if not sure
        trendline="ols",
        # Define the trendline as logarithmic, this will return curved functions
        trendline_options=dict(log_x=True)
    )
    # Print Figure
    fig.show()

# In[158]:


import plotly.express as px

for m in media:
    s_df = pd.merge(df[m], df_saturation_response[m], left_index=True, right_index=True)

    # For each media remove all the 0 spend values as Log(0) can't be calculated
    s_df[m + '_x'] = s_df[m + '_x'][s_df[m + '_x'] != 0]
    s_df[m + '_y'] = s_df[m + '_y'][s_df[m + '_y'] != 0]
    # saturation_df = saturation_df[(saturation_df[[media]] != 0).all(axis=1)]

    # Create a Scatter Plot
    fig = px.line(
        # Our temporary DF with no 0 values
        s_df,
        # Media channel's name
        x=m + '_x',
        # Our output variable (Sales or Conversion)
        y=m + '_y',
        # Specify the type of trendline you want to show, use OLS if not sure
        marker=True,
        # Define the trendline as logarithmic, this will return curved functions
        # trendline_options=dict(log_x=True)
    )
    # Print Figure
    fig.show()

# In[93]:


df_aggregated, df_all_decomp = decomposition_to_dataset(result, coefficenti_dict, columns_coef, spends_df, media,
                                                        'date', 'conversions', 'prediction')
save_dataset_to_csv(df_all_decomp, 'C:/Users/Federico/Downloads/treddom_ES_Timeseries.csv')

df_aggregated = pd.merge(how='left', left=df_aggregated, right=df_adstock_saturation, left_on='canale',
                         right_on='canale')
save_dataset_to_csv(df_aggregated, 'C:/Users/Federico/Downloads/treddom_ES_Decomposition.csv')
df_aggregated

# In[94]:


df_all_decomp

# In[95]:


# lower_bounds = [100, spends[1]*0.6, spends[2]*0.6, spends[3]*0.6, spends[4]*0.6, spends[5]*0.6, spends[6]*0.6,
# spends[7]*0.6, spends[8]*0.6]
# upper_bounds = [spends[0]*1.5, spends[1]*1.5, spends[2]*1.5, spends[3]*1.5, spends[4]*1.5, spends[5]*1.5, spends[6]*1.5,
# spends[7]*1.5, spends[8]*1.5]

lower_bounds = [100, 0, 0, 50, 20, 50, 0, 0, 0]
upper_bounds = [165, 60, 60, 165, 165, 165, 60, 60, 60]

budget_allocator_df = budget_allocator(result, 'date', media, all_features, '2022-02-28',
                                       model, spends, response_get_budget, lower_bounds, upper_bounds, 5000,
                                       df_aggregated, algoritm='GN_ISRES')
save_dataset_to_csv(budget_allocator_df, 'C:/Users/Federico/Downloads/treddom_ES_budget_allocator.csv')
budget_allocator_df

# In[ ]:


# In[ ]:







import pandas as pd
import nlopt
import numpy as np


def budget_allocator(df, name_date_column, medias, all_features, date_get_budget, model, spends, response_get_budget,
                     lower_bounds, upper_bounds, maxeval, df_aggregated, algoritm='LD_MMA'):
    def getVals(df, name_date_column, all_features, date_get_budget):
        # last_week = datetime.strftime(datetime.today() - timedelta(7), '%Y-%m-%d')

        full_row = df.loc[df[name_date_column] == date_get_budget]
        row = full_row[all_features].copy()

        return row

    def myFunc(x, grad=[]):
        data = {}

        for m in medias:
            data[m] = [x[medias.index(m)]]

        dic = pd.DataFrame.from_dict(data)

        new_df = getVals(df, name_date_column, all_features, date_get_budget).copy()

        for column in dic:
            new_df[column] = dic[column].iloc[0]

        return model.predict(new_df)[0]

    if algoritm == 'GN_DIRECT':
        opt = nlopt.opt(nlopt.GN_DIRECT, len(medias))
    elif algoritm == 'LD_SLSQP':
        opt = nlopt.opt(nlopt.LD_SLSQP, len(medias))
    elif algoritm == 'GN_ISRES':
        opt = nlopt.opt(nlopt.GN_ISRES, len(medias))
    elif algoritm == 'GN_AGS':
        opt = nlopt.opt(nlopt.GN_AGS, len(medias))
    elif algoritm == 'LD_COBYLA':
        opt = nlopt.opt(nlopt.LD_COBYLA, len(medias))
    elif algoritm == 'GN_CRS2_LM':
        opt = nlopt.opt(nlopt.GN_CRS2_LM, len(medias))
    elif algoritm == 'G_MLSL':
        opt = nlopt.opt(nlopt.G_MLSL, len(medias))
    elif algoritm == 'GD_STOGO':
        opt = nlopt.opt(nlopt.GD_STOGO, len(medias))
    elif algoritm == 'GN_ESCH':
        opt = nlopt.opt(nlopt.GN_ESCH, len(medias))
    elif algoritm == 'LN_BOBYQA':
        opt = nlopt.opt(nlopt.LN_BOBYQA, len(medias))
    elif algoritm == 'LN_NEWUOA':
        opt = nlopt.opt(nlopt.LN_NEWUOA, len(medias))
    elif algoritm == 'LD_CCSAQ':
        opt = nlopt.opt(nlopt.LD_CCSAQ, len(medias))
    elif algoritm == 'AUGLAG':
        opt = nlopt.opt(nlopt.AUGLAG, len(medias))
    else:
        opt = nlopt.opt(nlopt.LD_MMA, len(medias))

    lower_boundaries = np.multiply(lower_bounds, spends)
    upper_boundaries = np.multiply(upper_bounds, spends)

    opt.set_lower_bounds(lower_boundaries)
    opt.set_upper_bounds(upper_boundaries)

    opt.set_max_objective(myFunc)
    opt.add_inequality_constraint(lambda z, grad: sum(z) - np.sum(spends), 1e-8)

    # rate of improvement, below which we are done
    opt.set_xtol_rel(1e-14)
    opt.set_maxeval(maxeval)

    budget_spends = opt.optimize(spends)

    budget_allocator_df = pd.DataFrame()
    budget_allocator_df['canale'] = medias
    for index, row in budget_allocator_df.iterrows():
        budget_allocator_df.at[index, 'actual_spend'] = pow(spends[index], 1/df_aggregated.loc[index, 'saturation'])
        budget_allocator_df.at[index, 'optimal_spend'] = pow(budget_spends[index], 1/df_aggregated.loc[index, 'saturation'])
        budget_allocator_df.at[index, 'actual_response'] = df_aggregated.loc[index, 'xDecompAgg']
        budget_allocator_df.at[index, 'optimal_response'] = budget_spends[index] * df_aggregated.loc[index, 'coef']
        budget_allocator_df.at[index, 'actual_total_spend'] = np.sum(spends)
        budget_allocator_df.at[index, 'optimal_total_spend'] = np.sum(budget_spends)
        budget_allocator_df.at[index, 'actual_total_response'] = response_get_budget
        budget_allocator_df.at[index, 'optimal_total_response'] = opt.last_optimum_value()

    return budget_allocator_df


from sklearn.linear_model import BayesianRidge
from sklearn.model_selection import train_test_split
from cassandra.data.trasformations.trasformations import create_model
from cassandra.model.modelEvaluation.plot import show_nrmse, show_mape, show_rsquared


def bayesianRidge(df, X_columns, target, name_model, medias=[], organic=[], metric=['rsq', 'nrmse', 'mape'],
                  return_metric=False, size=0.2, positive=False):
    metrics_values = {}
    X = df[X_columns]
    y = df[target]

    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=size)

    if medias or organic:
        model = create_model(medias, organic, BayesianRidge(positive=positive))
    else:
        model = BayesianRidge(positive=positive)

    model.fit(X_train, y_train)

    # Ask the model to predict on X_test without having Y_test
    # This will give you exact predicted values

    # We can use our NRMSE and MAPE functions as well

    # Create new DF not to edit the original one
    result = df

    # Create a new column with predicted values
    if medias or organic:
        result['prediction'] = model.predict(result)
    else:
        result['prediction'] = model.predict(X)

    # Score returns the accuracy of the above prediction or R^2
    if 'rsq' in metric:
        rsq = show_rsquared(result[target], result['prediction'])
        if return_metric:
            metrics_values[name_model + '_rsq'] = rsq
        print(name_model, 'RSQ: ', rsq)

    # Get the NRMSE & MAPE values
    if 'nrmse' in metric:
        nrmse_val = show_nrmse(result[target], result['prediction'])
        if return_metric:
            metrics_values[name_model + '_nrmse'] = nrmse_val
        print(name_model, 'NRMSE: ', nrmse_val)

    if 'mape' in metric:
        mape_val = show_mape(result[target], result['prediction'])
        if return_metric:
            metrics_values[name_model + '_mape'] = mape_val
        print(name_model, 'MAPE: ', mape_val)

    if metrics_values:
        return result, model, metrics_values
    else:
        return result, model



from cassandra.model.linear import linear
from cassandra.model.logLinear import logLinear
from cassandra.model.ridge import ridge
from cassandra.model.logLog import logLog


def nestedModel(df, X_columns, y_column, model_regression='linear', X_trasformations_columns=[],
                model_regression_log_log='linear', ridge_number=0, metric=['rsq', 'nrmse', 'mape'],
                return_metric=False, size=0.2, positive=False):
    metrics_values = {}
    name_model = 'Nested Model for ' + y_column

    if model_regression == 'linear':
        if return_metric:
            result, model, metrics_values = linear(df, X_columns, y_column, name_model, metric=metric,
                                                   return_metric=return_metric, size=size, positive=positive)
        else:
            result, model = linear(df, X_columns, y_column, name_model, metric=metric, return_metric=return_metric, size=size, positive=positive)

    elif model_regression == 'logLinear':
        if return_metric:
            result, model, metrics_values = logLinear(df, X_columns, y_column, name_model, metric=metric,
                                                      return_metric=return_metric, size=size, positive=positive)
        else:
            result, model = logLinear(df, X_columns, y_column, name_model, metric=metric, return_metric=return_metric, size=size, positive=positive)

    elif model_regression == 'logLog':
        if return_metric:
            result, model, metrics_values = logLog(df, X_trasformations_columns, X_columns, y_column, name_model,
                                                   model_regression=model_regression_log_log, metric=metric,
                                                   return_metric=return_metric, size=size, positive=positive)
        else:
            result, model = logLog(df, X_trasformations_columns, X_columns, y_column, name_model,
                                   model_regression=model_regression_log_log, metric=metric,
                                   return_metric=return_metric, size=size, positive=positive)

    elif model_regression == 'ridge':
        if return_metric:
            result, model, metrics_values = ridge(df, X_columns, y_column, name_model, ridge_number=ridge_number,
                                                  metric=metric,
                                                  return_metric=return_metric, size=size, positive=positive)
        else:
            result, model = ridge(df, X_columns, y_column, name_model, ridge_number=ridge_number, metric=metric,
                                  return_metric=return_metric, size=size, positive=positive)

    if metrics_values:
        return result, model, metrics_values
    else:
        return result, model


def mainNestedModel(sub_df, sub_X_columns, sub_y_column, X_columns, y_column, sub_model_regression='linear',
                    sub_X_trasformations_columns=[],
                    sub_model_regression_log_log='linear', sub_ridge_number=0, sub_metric=['rsq', 'nrmse', 'mape'],
                    return_sub_metric=False, sub_size=0.2, sub_positive=False, model_regression='linear',
                    X_trasformations_columns=[], model_regression_log_log='linear', ridge_number=0,
                    metric=['rsq', 'nrmse', 'mape'], return_metric=False, size=0.2, positive=False):
    metrics_values = {}
    if return_sub_metric == True:
        sub_result, sub_model, sub_metric_values = nestedModel(sub_df, sub_X_columns, sub_y_column,
                                                               sub_model_regression, sub_X_trasformations_columns,
                                                               sub_model_regression_log_log, sub_ridge_number,
                                                               sub_metric, return_sub_metric, sub_size=sub_size, sub_positive=sub_positive)
    else:
        sub_result, sub_model = nestedModel(sub_df, sub_X_columns, sub_y_column, sub_model_regression,
                                            sub_X_trasformations_columns,
                                            sub_model_regression_log_log, sub_ridge_number, sub_metric, sub_size=sub_size, sub_positive=sub_positive)

    df = sub_result.copy()

    df[sub_y_column] = sub_result['prediction'].copy()

    name_model = model_regression + ' for ' + y_column

    if model_regression == 'linear':
        if return_metric:
            result, model, metrics_values = linear(df, X_columns, y_column, name_model, metric=metric,
                                                   return_metric=return_metric, size=size, positive=positive)
        else:
            result, model = linear(df, X_columns, y_column, name_model, metric=metric, return_metric=return_metric, size=size, positive=positive)

    elif model_regression == 'logLinear':
        if return_metric:
            result, model, metrics_values = logLinear(df, X_columns, y_column, name_model, metric=metric,
                                                      return_metric=return_metric, size=size, positive=positive)
        else:
            result, model = logLinear(df, X_columns, y_column, name_model, metric=metric, return_metric=return_metric, size=size, positive=positive)

    elif model_regression == 'logLog':
        if return_metric:
            result, model, metrics_values = logLog(df, X_trasformations_columns, X_columns, y_column, name_model,
                                                   model_regression=model_regression_log_log, metric=metric,
                                                   return_metric=return_metric, size=size, positive=positive)
        else:
            result, model = logLog(df, X_trasformations_columns, X_columns, y_column, name_model,
                                   model_regression=model_regression_log_log, metric=metric,
                                   return_metric=return_metric, size=size, positive=positive)

    elif model_regression == 'ridge':
        if return_metric:
            result, model, metrics_values = ridge(df, X_columns, y_column, name_model, ridge_number=ridge_number,
                                                  metric=metric, return_metric=return_metric, size=size, positive=positive)
        else:
            result, model = ridge(df, X_columns, y_column, name_model, ridge_number=ridge_number, metric=metric,
                                  return_metric=return_metric, size=size, positive=positive)

    if metrics_values:
        return result, model, metrics_values, sub_result, sub_model, sub_metric_values
    else:
        return result, model, sub_result, sub_model



from datetime import datetime, timedelta


def getBudget(df, name_date_column, name_target_column, medias, date_get_budget):
    full_row = df.loc[df[name_date_column] == date_get_budget]

    spents = []

    for m in medias:
        spents.append(full_row[m].values[0])

    response = full_row[name_target_column].values[0]
    return spents, response


def getVals(df, name_date_column, all_features, date_get_budget):

    full_row = df.loc[df[name_date_column] == date_get_budget]
    row = full_row[all_features].copy()

    return row



import statsmodels.api as sm


def ols(X, y, constant=True):
    features = X.copy()

    if constant:
        features = sm.add_constant(features)

    model = sm.OLS(y, features)

    result = model.fit()
    print(result.summary())

    return result, model
